package com.sun.java.util.jar.pack;

import java.util.ArrayList;
import java.util.Arrays;
import java.io.IOException;
import java.io.FilterInputStream;
import java.io.InputStream;
import java.util.Map;
import java.io.DataInputStream;

class ClassReader
{
    int verbose;
    Package pkg;
    Package.Class cls;
    long inPos;
    long constantPoolLimit;
    DataInputStream in;
    Map<Attribute.Layout, Attribute> attrDefs;
    Map<Attribute.Layout, String> attrCommands;
    String unknownAttrCommand;
    private static final ConstantPool.Entry INVALID_ENTRY;
    boolean haveUnresolvedEntry;
    
    ClassReader(Package.Class cls, final InputStream inputStream) throws IOException {
        this.constantPoolLimit = -1L;
        this.unknownAttrCommand = "error";
        this.pkg = cls.getPackage();
        this.cls = cls;
        this.verbose = this.pkg.verbose;
        this.in = new DataInputStream(new FilterInputStream(inputStream) {
            @Override
            public int read(byte[] b, int off, int len) throws IOException {
                int read = super.read(b, off, len);
                if (read >= 0) {
                    ClassReader this$0 = ClassReader.this;
                    this$0.inPos += (long)read;
                }
                return read;
            }
            
            @Override
            public int read() throws IOException {
                int read = super.read();
                if (read >= 0) {
                    ClassReader this$0 = ClassReader.this;
                    ++this$0.inPos;
                }
                return read;
            }
            
            @Override
            public long skip(long n) throws IOException {
                long skip = super.skip(n);
                if (skip >= 0L) {
                    ClassReader this$0 = ClassReader.this;
                    this$0.inPos += skip;
                }
                return skip;
            }
        });
    }
    
    public void setAttrDefs(Map<Attribute.Layout, Attribute> attrDefs) {
        this.attrDefs = attrDefs;
    }
    
    public void setAttrCommands(Map<Attribute.Layout, String> attrCommands) {
        this.attrCommands = attrCommands;
    }
    
    private void skip(int i, String str) throws IOException {
        Utils.log.warning("skipping " + i + " bytes of " + str);
        long n = 0L;
        while (n < (long)i) {
            long skip = this.in.skip((long)i - n);
            assert skip > 0L;
            n += skip;
        }
        assert n == (long)i;
        return;
    }
    
    private int readUnsignedShort() throws IOException {
        return this.in.readUnsignedShort();
    }
    
    private int readInt() throws IOException {
        return this.in.readInt();
    }
    
    private ConstantPool.Entry readRef() throws IOException {
        int unsignedShort = this.in.readUnsignedShort();
        return (unsignedShort == 0) ? null : this.cls.cpMap[unsignedShort];
    }
    
    private ConstantPool.Entry readRef(byte b) throws IOException {
        ConstantPool.Entry ref = this.readRef();
        assert !(ref instanceof ClassReader.UnresolvedEntry);
        this.checkTag(ref, b);
        return ref;
    }
    
    private ConstantPool.Entry checkValid(ConstantPool.Entry entry) {
        if (entry == ClassReader.INVALID_ENTRY) {
            throw new IllegalStateException("Invalid constant pool reference");
        }
        return entry;
    }
    
    private ConstantPool.Entry checkTag(ConstantPool.Entry entry, byte b) throws ClassReader.ClassFormatException {
        if (entry == null || !entry.tagMatches((int)b)) {
            throw new ClassReader.ClassFormatException("Bad constant, expected type=" + ConstantPool.tagName((int)b) + " got " + ((entry == null) ? "null CP index" : ("type=" + ConstantPool.tagName((int)entry.tag))) + ", in File: " + this.cls.file.nameString + ((this.inPos == this.constantPoolLimit) ? " in constant pool" : (" at pos: " + this.inPos)));
        }
        return entry;
    }
    
    private ConstantPool.Entry checkTag(ConstantPool.Entry entry, byte b, boolean b2) throws ClassReader.ClassFormatException {
        return (b2 && entry == null) ? null : this.checkTag(entry, b);
    }
    
    private ConstantPool.Entry readRefOrNull(byte b) throws IOException {
        ConstantPool.Entry ref = this.readRef();
        this.checkTag(ref, b, true);
        return ref;
    }
    
    private ConstantPool.Utf8Entry readUtf8Ref() throws IOException {
        return (ConstantPool.Utf8Entry)this.readRef((byte)1);
    }
    
    private ConstantPool.ClassEntry readClassRef() throws IOException {
        return (ConstantPool.ClassEntry)this.readRef((byte)7);
    }
    
    private ConstantPool.ClassEntry readClassRefOrNull() throws IOException {
        return (ConstantPool.ClassEntry)this.readRefOrNull((byte)7);
    }
    
    private ConstantPool.SignatureEntry readSignatureRef() throws IOException {
        ConstantPool.Entry ref = this.readRef((byte)13);
        return (ref != null && ref.getTag() == 1) ? ConstantPool.getSignatureEntry(ref.stringValue()) : ((ConstantPool.SignatureEntry)ref);
    }
    
    void read() throws IOException {
        boolean n = false;
        try {
            this.readMagicNumbers();
            this.readConstantPool();
            this.readHeader();
            this.readMembers(false);
            this.readMembers(true);
            this.readAttributes(0, this.cls);
            this.fixUnresolvedEntries();
            this.cls.finishReading();
            assert 0 >= this.in.read(new byte[1]);
            n = true;
        }
        finally {
            if (n == 0 && this.verbose > 0) {
                Utils.log.warning("Erroneous data at input offset " + this.inPos + " of " + this.cls.file);
            }
        }
    }
    
    void readMagicNumbers() throws IOException {
        this.cls.magic = this.in.readInt();
        if (this.cls.magic != -889275714) {
            throw new Attribute.FormatException("Bad magic number in class file " + Integer.toHexString(this.cls.magic), 0, "magic-number", "pass");
        }
        this.cls.version = Package.Version.of((int)(short)this.readUnsignedShort(), (int)(short)this.readUnsignedShort());
        String checkVersion = this.checkVersion(this.cls.version);
        if (checkVersion != null) {
            throw new Attribute.FormatException("classfile version too " + checkVersion + ": " + this.cls.version + " in " + this.cls.file, 0, "version", "pass");
        }
        return;
    }
    
    private String checkVersion(Package.Version version) {
        int major = version.major;
        int minor = version.minor;
        if (major < this.pkg.minClassVersion.major || (major == this.pkg.minClassVersion.major && minor < this.pkg.minClassVersion.minor)) {
            return "small";
        }
        if (major > this.pkg.maxClassVersion.major || (major == this.pkg.maxClassVersion.major && minor > this.pkg.maxClassVersion.minor)) {
            return "large";
        }
        return null;
    }
    
    void readConstantPool() throws IOException {
        int unsignedShort = this.in.readUnsignedShort();
        int[] array = new int[unsignedShort * 4];
        int i = 0;
        ConstantPool.Entry[] cpMap = new ConstantPool.Entry[unsignedShort];
        cpMap[0] = ClassReader.INVALID_ENTRY;
        for (int j = 1; j < unsignedShort; ++j) {
            int byte1 = this.in.readByte();
            switch (byte1) {
                case 1: {
                    cpMap[j] = ConstantPool.getUtf8Entry(this.in.readUTF());
                    break;
                }
                case 3: {
                    cpMap[j] = ConstantPool.getLiteralEntry(Integer.valueOf(this.in.readInt()));
                    break;
                }
                case 4: {
                    cpMap[j] = ConstantPool.getLiteralEntry(Float.valueOf(this.in.readFloat()));
                    break;
                }
                case 5: {
                    cpMap[j] = ConstantPool.getLiteralEntry(Long.valueOf(this.in.readLong()));
                    cpMap[++j] = ClassReader.INVALID_ENTRY;
                    break;
                }
                case 6: {
                    cpMap[j] = ConstantPool.getLiteralEntry(Double.valueOf(this.in.readDouble()));
                    cpMap[++j] = ClassReader.INVALID_ENTRY;
                    break;
                }
                case 7:
                case 8:
                case 16: {
                    array[i++] = j;
                    array[i++] = byte1;
                    array[i++] = this.in.readUnsignedShort();
                    array[i++] = -1;
                    break;
                }
                case 9:
                case 10:
                case 11:
                case 12: {
                    array[i++] = j;
                    array[i++] = byte1;
                    array[i++] = this.in.readUnsignedShort();
                    array[i++] = this.in.readUnsignedShort();
                    break;
                }
                case 18: {
                    array[i++] = j;
                    array[i++] = byte1;
                    array[i++] = (-1 ^ this.in.readUnsignedShort());
                    array[i++] = this.in.readUnsignedShort();
                    break;
                }
                case 15: {
                    array[i++] = j;
                    array[i++] = byte1;
                    array[i++] = (-1 ^ this.in.readUnsignedByte());
                    array[i++] = this.in.readUnsignedShort();
                    break;
                }
                default: {
                    throw new ClassReader.ClassFormatException("Bad constant pool tag " + byte1 + " in File: " + this.cls.file.nameString + " at pos: " + this.inPos);
                }
            }
        }
        this.constantPoolLimit = this.inPos;
        while (i > 0) {
            if (this.verbose > 3) {
                Utils.log.fine("CP fixups [" + i / 4 + "]");
            }
            i = 0;
            int k = 0;
            while (k < i) {
                int l = array[k++];
                int n = array[k++];
                int m = array[k++];
                int i2 = array[k++];
                if (this.verbose > 3) {
                    Utils.log.fine("  cp[" + l + "] = " + ConstantPool.tagName(n) + "{" + m + "," + i2 + "}");
                }
                if ((m >= 0 && this.checkValid(cpMap[m]) == null) || (i2 >= 0 && this.checkValid(cpMap[i2]) == null)) {
                    array[i++] = l;
                    array[i++] = n;
                    array[i++] = m;
                    array[i++] = i2;
                }
                else {
                    switch (n) {
                        case 7: {
                            cpMap[l] = ConstantPool.getClassEntry(cpMap[m].stringValue());
                            continue;
                        }
                        case 8: {
                            cpMap[l] = ConstantPool.getStringEntry(cpMap[m].stringValue());
                            continue;
                        }
                        case 9:
                        case 10:
                        case 11: {
                            cpMap[l] = ConstantPool.getMemberEntry((byte)n, (ConstantPool.ClassEntry)this.checkTag(cpMap[m], (byte)7), (ConstantPool.DescriptorEntry)this.checkTag(cpMap[i2], (byte)12));
                            continue;
                        }
                        case 12: {
                            cpMap[l] = ConstantPool.getDescriptorEntry((ConstantPool.Utf8Entry)this.checkTag(cpMap[m], (byte)1), (ConstantPool.Utf8Entry)this.checkTag(cpMap[i2], (byte)13));
                            continue;
                        }
                        case 16: {
                            cpMap[l] = ConstantPool.getMethodTypeEntry((ConstantPool.Utf8Entry)this.checkTag(cpMap[m], (byte)13));
                            continue;
                        }
                        case 15: {
                            cpMap[l] = ConstantPool.getMethodHandleEntry((byte)(-1 ^ m), (ConstantPool.MemberEntry)this.checkTag(cpMap[i2], (byte)52));
                            continue;
                        }
                        case 18: {
                            cpMap[l] = new ClassReader.UnresolvedEntry((byte)n, new Object[] { Integer.valueOf(-1 ^ m), (ConstantPool.Entry)this.checkTag(cpMap[i2], (byte)12) });
                            continue;
                        }
                        default: {
                            assert false;
                            continue;
                        }
                    }
                }
            }
            assert i < i;
            continue;
        }
        this.cls.cpMap = cpMap;
    }
    
    private void fixUnresolvedEntries() {
        if (!this.haveUnresolvedEntry) {
            return;
        }
        ConstantPool.Entry[] cpMap = this.cls.getCPMap();
        for (int i = 0; i < cpMap.length; ++i) {
            ConstantPool.Entry entry = cpMap[i];
            if ((entry instanceof ClassReader.UnresolvedEntry)) {
                ConstantPool.Entry entry2 = cpMap[i] = ((ClassReader.UnresolvedEntry)entry).resolve();
                assert !(entry2 instanceof ClassReader.UnresolvedEntry);
            }
        }
        this.haveUnresolvedEntry = false;
        return;
    }
    
    void readHeader() throws IOException {
        this.cls.flags = this.readUnsignedShort();
        this.cls.thisClass = this.readClassRef();
        this.cls.superClass = this.readClassRefOrNull();
        int unsignedShort = this.readUnsignedShort();
        this.cls.interfaces = new ConstantPool.ClassEntry[unsignedShort];
        for (int i = 0; i < unsignedShort; ++i) {
            this.cls.interfaces[i] = this.readClassRef();
        }
    }
    
    void readMembers(boolean b) throws IOException {
        for (int unsignedShort = this.readUnsignedShort(), i = 0; i < unsignedShort; ++i) {
            this.readMember(b);
        }
    }
    
    void readMember(boolean b) throws IOException {
        int unsignedShort = this.readUnsignedShort();
        ConstantPool.DescriptorEntry descriptorEntry = ConstantPool.getDescriptorEntry(this.readUtf8Ref(), this.readSignatureRef());
        Package.Class.Field field;
        if (!b) {
            field = this.cls.new Package.Class.Field(unsignedShort, descriptorEntry);
        }
        else {
            field = this.cls.new Package.Class.Method(unsignedShort, descriptorEntry);
        }
        this.readAttributes(b ? 2 : 1, field);
    }
    
    void readAttributes(int n, Attribute.Holder obj) throws IOException {
        int unsignedShort = this.readUnsignedShort();
        if (unsignedShort == 0) {
            return;
        }
        if (this.verbose > 3) {
            Utils.log.fine("readAttributes " + obj + " [" + unsignedShort + "]");
        }
        for (int i = 0; i < unsignedShort; ++i) {
            String stringValue = this.readUtf8Ref().stringValue();
            int int1 = this.readInt();
            if (this.attrCommands != null) {
                String s = (String)this.attrCommands.get(Attribute.keyForLookup(n, stringValue));
                if (s != null) {
                    int n2 = -1;
                    switch (s.hashCode()) {
                        case 3433489: {
                            if (s.equals("pass")) {
                                n2 = 0;
                                break;
                            }
                            break;
                        }
                        case 96784904: {
                            if (s.equals("error")) {
                                n2 = 1;
                                break;
                            }
                            break;
                        }
                        case 109773592: {
                            if (s.equals("strip")) {
                                n2 = 2;
                                break;
                            }
                            break;
                        }
                    }
                    switch (n2) {
                        case 0: {
                            throw new Attribute.FormatException("passing attribute bitwise in " + obj, n, stringValue, s);
                        }
                        case 1: {
                            throw new Attribute.FormatException("attribute not allowed in " + obj, n, stringValue, s);
                        }
                        case 2: {
                            this.skip(int1, stringValue + " attribute in " + obj);
                            continue;
                        }
                        default: {
                            break;
                        }
                    }
                }
            }
            Attribute attribute = Attribute.lookup(Package.attrDefs, n, stringValue);
            if (this.verbose > 4 && attribute != null) {
                Utils.log.fine("pkg_attribute_lookup " + stringValue + " = " + attribute);
            }
            if (attribute == null) {
                attribute = Attribute.lookup(this.attrDefs, n, stringValue);
                if (this.verbose > 4 && attribute != null) {
                    Utils.log.fine("this " + stringValue + " = " + attribute);
                }
            }
            if (attribute == null) {
                attribute = Attribute.lookup(null, n, stringValue);
                if (this.verbose > 4 && attribute != null) {
                    Utils.log.fine("null_attribute_lookup " + stringValue + " = " + attribute);
                }
            }
            if (attribute == null && int1 == 0) {
                attribute = Attribute.find(n, stringValue, "");
            }
            int n3 = n == 3 && (stringValue.equals("StackMap") || stringValue.equals("StackMapX"));
            if (!(n3 == 0)) {
                Code code = (Code)obj;
                if (code.max_stack >= 65536 || code.max_locals >= 65536 || code.getLength() >= 65536 || stringValue.endsWith("X")) {
                    attribute = null;
                }
            }
            if (attribute == null) {
                if (!(n3 == 0)) {
                    throw new Attribute.FormatException("unsupported StackMap variant in " + obj, n, stringValue, "pass");
                }
                if ("strip".equals(this.unknownAttrCommand)) {
                    this.skip(int1, "unknown " + stringValue + " attribute in " + obj);
                }
                else {
                    throw new Attribute.FormatException(" is unknown attribute in class " + obj, n, stringValue, this.unknownAttrCommand);
                }
            }
            else {
                long inPos = this.inPos;
                if (attribute.layout() == Package.attrCodeEmpty) {
                    Package.Class.Method method = (Package.Class.Method)obj;
                    method.code = new Code(method);
                    try {
                        this.readCode(method.code);
                    }
                    catch (Instruction.FormatException ex) {
                        throw new ClassReader.ClassFormatException(ex.getMessage() + " in " + obj, ex);
                    }
                    assert (long)int1 == this.inPos - inPos;
                }
                else if (attribute.layout() == Package.attrBootstrapMethodsEmpty) {
                    assert obj == this.cls;
                    this.readBootstrapMethods(this.cls);
                    assert (long)int1 == this.inPos - inPos;
                    continue;
                }
                else if (attribute.layout() == Package.attrInnerClassesEmpty) {
                    assert obj == this.cls;
                    this.readInnerClasses(this.cls);
                    assert (long)int1 == this.inPos - inPos;
                }
                else if (int1 > 0) {
                    byte[] b = new byte[int1];
                    this.in.readFully(b);
                    attribute = attribute.addContent(b);
                }
                if (attribute.size() == 0 && !attribute.layout().isEmpty()) {
                    throw new ClassReader.ClassFormatException(stringValue + ": attribute length cannot be zero, in " + obj);
                }
                obj.addAttribute(attribute);
                if (this.verbose > 2) {
                    Utils.log.fine("read " + attribute);
                }
            }
        }
        return;
    }
    
    void readCode(Code code) throws IOException {
        code.max_stack = this.readUnsignedShort();
        code.max_locals = this.readUnsignedShort();
        code.bytes = new byte[this.readInt()];
        this.in.readFully(code.bytes);
        Instruction.opcodeChecker(code.bytes, this.cls.getCPMap(), this.cls.version);
        int unsignedShort = this.readUnsignedShort();
        code.setHandlerCount(unsignedShort);
        for (int i = 0; i < unsignedShort; ++i) {
            code.handler_start[i] = this.readUnsignedShort();
            code.handler_end[i] = this.readUnsignedShort();
            code.handler_catch[i] = this.readUnsignedShort();
            code.handler_class[i] = this.readClassRefOrNull();
        }
        this.readAttributes(3, code);
    }
    
    void readBootstrapMethods(Package.Class class1) throws IOException {
        ConstantPool.BootstrapMethodEntry[] a = new ConstantPool.BootstrapMethodEntry[this.readUnsignedShort()];
        ConstantPool.MethodHandleEntry methodHandleEntry;
        ConstantPool.Entry[] array;
        for (int i = 0; i < a.length; a[i] = ConstantPool.getBootstrapMethodEntry(methodHandleEntry, array), ++i) {
            methodHandleEntry = (ConstantPool.MethodHandleEntry)this.readRef((byte)15);
            array = new ConstantPool.Entry[this.readUnsignedShort()];
            for (int j = 0; j < array.length; ++j) {
                array[j] = this.readRef((byte)51);
            }
        }
        class1.setBootstrapMethods(Arrays.asList(a));
    }
    
    void readInnerClasses(Package.Class class1) throws IOException {
        int unsignedShort = this.readUnsignedShort();
        ArrayList innerClasses = new ArrayList(unsignedShort);
        for (int i = 0; i < unsignedShort; ++i) {
            innerClasses.add(new Package.InnerClass(this.readClassRef(), this.readClassRefOrNull(), (ConstantPool.Utf8Entry)this.readRefOrNull((byte)1), this.readUnsignedShort()));
        }
        class1.innerClasses = innerClasses;
    }
    
    static {
        INVALID_ENTRY = new ConstantPool.Entry((byte)(-1)) {
            @Override
            public boolean equals(Object o) {
                throw new IllegalStateException("Should not call this");
            }
            
            @Override
            protected int computeValueHash() {
                throw new IllegalStateException("Should not call this");
            }
            
            @Override
            public int compareTo(Object o) {
                throw new IllegalStateException("Should not call this");
            }
            
            @Override
            public String stringValue() {
                throw new IllegalStateException("Should not call this");
            }
        };
    }
    
    private class UnresolvedEntry extends ConstantPool.Entry
    {
        final Object[] refsOrIndexes;
        
        UnresolvedEntry(byte b, Object... refsOrIndexes) {
            super(b);
            this.refsOrIndexes = refsOrIndexes;
            ClassReader.this.haveUnresolvedEntry = true;
        }
        
        ConstantPool.Entry resolve() {
            Package.Class cls = ClassReader.this.cls;
            switch (this.tag) {
                case 18: {
                    return ConstantPool.getInvokeDynamicEntry((ConstantPool.BootstrapMethodEntry)cls.bootstrapMethods.get((int)this.refsOrIndexes[0]), (ConstantPool.DescriptorEntry)this.refsOrIndexes[1]);
                }
                default: {
                    throw new AssertionError();
                }
            }
        }
        
        private void unresolved() {
            throw new RuntimeException("unresolved entry has no string");
        }
        
        @Override
        public int compareTo(Object o) {
            this.unresolved();
            return 0;
        }
        
        @Override
        public boolean equals(Object o) {
            this.unresolved();
            return false;
        }
        
        @Override
        protected int computeValueHash() {
            this.unresolved();
            return 0;
        }
        
        @Override
        public String stringValue() {
            this.unresolved();
            return this.toString();
        }
        
        @Override
        public String toString() {
            return "(unresolved " + ConstantPool.tagName((int)this.tag) + ")";
        }
    }
    
    static class ClassFormatException extends IOException
    {
        private static final long serialVersionUID = -3564121733989501833L;
        
        public ClassFormatException(String message) {
            super(message);
        }
        
        public ClassFormatException(String message, Throwable cause) {
            super(message, cause);
        }
    }
}
